"""
App
---

Das `app`-Modul mit der Klasse `Opus` ist der Kern von `opus`.
"""

import csv
from io import StringIO
import logging
import subprocess

from opus import server, targeting, backend
from opus.exceptions import *


class Opus:
    """Die Klasse `Opus` stellt die API für alle Frontends von `opus` bereit.

    Beispiel:

    .. code-block:: python

        app = Opus(url, username, password)

        # get a list of all known clients (instances of Client)
        clients = app.clients()
        # get a list of all installable products
        products = app.products()

    """

    def __init__(self, url, username, password):
        self.__backend = backend.Backend(url, username, password)
        logging.debug("built Opus instance")

    def clients(self):
        """Get all clients.

        :return: dictionary with client id:data; data is a dictionary
            with `products`:[(productId, productVersion, packageVersion,
            installationStatus, lastAction, actionResult)]
        """

        logging.debug("app.Opus.clients: getting clients")

        return self.__backend.clients

    def is_fresh(self, client_json):
        """Check if client is a fresh installed machine.

        :param dict client_json: client data (dict)
        :return: True if fresh, False otherwise
        """

        found_opsi_client_agent = False
        found_opsi_script = False
        found_swaudit = False

        if len(client_json["products"])==0:
            return True

        if len(client_json["products"])==2:
            for product in client_json["products"]:
                if product[0] == "opsi-client-agent":
                    found_opsi_client_agent = True
                elif product[0] == "opsi-script":
                    found_opsi_script = True

            return (found_opsi_client_agent and found_opsi_script)
        elif len(client_json["products"])==3:
            for product in client_json["products"]:
                if product[0] == "opsi-client-agent":
                    found_opsi_client_agent = True
                elif product[0] == "opsi-script":
                    found_opsi_script = True
                elif product[0] == "swaudit":
                    found_swaudit = True

            return (found_opsi_client_agent and found_opsi_script and found_swaudit)
        else:
            return False

    def clientGroups(self, group_id=""):
        """Get all client group names.

        :return: dictionary with group ids and descriptions
        """

        logging.debug("getting client groups for '%s'" % group_id)

        return self.__backend.clientGroups(group_id)

    def clientGroupMembers(self, group_id):
        """Get all members of a client group.

        :param str group_id: group id
        :return: list with client ids
        """

        logging.debug("app:getting client group members for '%s'" % group_id)

        return self.__backend.clientGroupMembers(group_id)

    def addClientsToGroup(self, client_ids, group_id):
        """Add Client to a group.

        :param list client_ids: list with client ids
        :param str group_id: group id
        """

        logging.debug("adding client(s) '%s' to group '%s'" % (client_ids, group_id))

        return self.__backend.addClientsToGroup(client_ids, group_id)

    def products(self):
        """Get a list all available products in the depot.

        Example for product data:

        .. code-block:: python

            product_data = {
                'name': 'Microsoft Visual C++ 2017', 'licenseRequired': False,
                'setupScript': 'setup3264.opsiscript', 'uninstallScript': '',
                'updateScript': '', 'alwaysScript': '', 'onceScript': '',
                'customScript': '', 'userLoginScript': '', 'priority': 0,
                'description': 'Microsoft Visual C++ 2017 redistributable',
                'advice': '', 'changelog': '', 'productClassIds': [],
                'windowsSoftwareIds': [], 'id': 'visualcpp-2017',
                'productVersion': '14.12.25810.0', 'packageVersion': '1',
                'type': 'LocalbootProduct',
                'ident': 'visualcpp-2017;14.12.25810.0;1'
                }


        :return: dictionary with product id: product data
        """

        logging.debug("app.Opus.products")

        return self.__backend.products

    def productGroups(self):
        """Get a list with known product groups.

        :return: dictionary with product group ids and description
        """

        logging.debug("getting product groups")

        return self.__backend.productGroups()

    def updatable(self, product_triple, products_in_depot):
        """Check if a product can be updated.

        :param tuple product_triple: triple of product data
        :return: new version (package version incl.) if available, otherwise False
        """

        logging.debug("looking for updates on %s" % product_triple[0])

        # compare product version and package version
        id_ = product_triple[0]
        # hardcoded, disabled windomain package
        if id_ == 'windomain':
            return False
        version = product_triple[1]
        package_version = product_triple[2]
        if id_ in products_in_depot:
            if gt(products_in_depot[id_], product_triple):
                return "%s-%s" % (products_in_depot[id_]["productVersion"],
                     products_in_depot[id_]["packageVersion"])
            else:
                return False
        else:
            logging.error(f"{id_} is not in depot!")

    def failedClients(self, client_ids):
        """Get list of failed clients.

        A failed client is a client with a failed un-/install action.

        :param list client_ids: client ids
        :return: dictionary with client id:[failed products]
        """

        logging.debug("getting failed clients")

        clients = self.__backend.clients

        failed_clients = {}

        for client_id in client_ids:
            products = clients[client_id]["products"]

            for product in products:
                if self.is_failed(product):
                    if not client_id in failed_clients:
                        failed_clients[client_id] = []
                    failed_clients[client_id].append(product[0])

        return failed_clients

    def is_failed(self, product_tuple):
        """Get a list with failed products on a client.

        :param tuple product_tuple: product data as a tupel
        :return: True if failed, False otherwise
        """

        installationStatus = product_tuple[3]
        lastAction = product_tuple[4]
        actionResult = product_tuple[5]

        if (lastAction=="setup" or lastAction=="update") and actionResult=="failed":
            return True
        elif lastAction=="uninstall" and actionResult=="failed":
            return True
        else:
            return False

    def productGroupMembers(self, group_id):
        """Get a list of products of a product group

        :param str group_id: group id
        :return: list with product ids and version
        """

        logging.debug("getting products group members for '%s'" % group_id)

        return self.__backend.productGroupMembers(group_id)

    def evaluate_targets(self, target_string):
        """Filter clients with given string and return the resulting list.
        der Liste der bekannten Clients.

        :param str target_string: filter string
        :return: list with clients (dictionaries)
        """

        return targeting.evaluate(target_string, self.clients(), self)

    def install(self, clients, packages, force, now):
        """Install given package(s) on client(s).

        Installation will be done if:

        * package is not installed
        * older version of the package installed

        :param dict clients: client dict with name:data
        :param list packages: list with package ids
        :param bool force: force installation even if package is installed
        :param bool now: trigger ondemand to install immediately
        """

        logging.debug("app.Opus.install")

        # first check if all packages are available on depot
        logging.debug("app.Opus.install: checking available packages")
        for p in packages:
            if p not in self.products():
                raise NoSuchProduct(p)

        for client, data in clients.items():
            logging.debug("app.Opus.install: client %s" % client)

            packages_on_client = data["products"]

            # iterate over the given packages list
            for package in packages:
                logging.debug("app.Opus.install: checking for package %s" % package)
                # with `force` simply install the packages
                if force:
                    self.__backend.install(client, package)
                else:
                    # get triple for given package on depot
                    p = self.products()[package]

                    found_on_client = False
                    for p_o_c in packages_on_client:
                        if p["id"] != p_o_c[0]:
                            # skip because this is not the right package
                            continue
                        logging.debug("app.Opus.install: checking against installed package %s" % p_o_c[0])
                        # install only newer versions
                        if gt(p, p_o_c):
                            logging.debug("app.Opus.install: %s is newer than %s -> installing" % (p, p_o_c))
                            self.__backend.install(client, package)
                            found_on_client = True
                            break
                        else:
                            logging.debug("app.Opus.install: %s versions are equal -> not installing" % (p_o_c[0]))
                            found_on_client = True
                            break

                    # if this package in not installed on client, install it
                    if not found_on_client:
                        logging.debug("app.Opus.install: %s not found on client -> installing" % (p["id"]))
                        self.__backend.install(client, package)

            if now:
                self.__backend.ondemand(client)

    def uninstall(self, clients, packages, now):
        """Remove package(s) from client(s).

        Deinstallation only takes place if the package is installed.

        :param dict clients: clients with name:data
        :param list packages: list with package ids
        :param bool now: trigger ondemand to uninstall immediately
        """

        logging.debug("app.Opus.uninstall")

        # first check if all packages are available on depot
        logging.debug("app.Opus.uninstall: checking available packages")
        for p in packages:
            if not p in self.products():
                raise NoSuchProduct(p)

        for client, data in clients.items():
            logging.debug("app.Opus.uninstall: working on %s" % client)

            packages_on_client = data["products"]

            # iterate over the given packages list
            for package in packages:
                logging.debug("app.Opus.uninstall: checking for package %s" % package)

                # get triple for given package on depot
                p = self.products()[p]

                found_on_client = False
                for p_o_c in packages_on_client:
                    if p["id"] != p_o_c[0]:
                        # skip because this is not the right package
                        continue
                    else:
                        logging.debug("app.Opus.uninstall: found package %s -> uninstalling" % p["id"])
                        self.__backend.uninstall(client, package)

            if now:
                self.__backend.ondemand(client)

    def update(self, clients, packages, now):
        """Update package(s) on client(s).

        Warning: `windomain` can break AD joins, if Rejoin is forbidden!

        :param dict clients: dict with name:data
        :param list packages: list with package names.  If this is empty, then
            all possible updates will be installed.
        :param bool now: trigger ondemand on clients
        """

        logging.debug("app.Opus.update")

        # first check if all packages are available on depot
        logging.debug("app.Opus.update: checking available packages")

        # this works only for a given list; if you update all packages
        # packages will be empty
        # this errors out soon enough before we start working on every client
        for p in packages:
            if not p in self.products():
                # this may be caused by a previously available package
                # that is not part of the depot anymore!
                raise NoSuchProduct(p)

        for client, data in clients.items():
            # now get all packages currently installed on the client
            # flatten to simple list
            packages_on_client = [ p[0] for p in data["products"] ]
            logging.debug("app.Opus.update: found %s installed packages on client %s" % (packages_on_client, client))
            # check for missing packages in depot
            for poc in packages_on_client:
                if poc not in self.products():
                    logging.error(f"app.Opus.update: {poc} not in depot! Skipping it.")
                    packages_on_client.remove(poc)
                elif poc == 'windomain':
                    # remove possible dangerous package
                    packages_on_client.remove(poc)

            # which of the given packages are installed? or update all installed packages?
            if len(packages) == 0:
                # remove missing packages to allow updating the rest
                installed_packages = packages_on_client
            else:
                installed_packages = [ p if p in packages_on_client else '' for p in packages ]

            logging.debug("app.Opus.update: found %s from given list" % installed_packages)
            if len(installed_packages) > 0:
                # use the "update" mechanism of our install method
                #logging.debug("app.Opus.update: would install %s" % installed_packages)
                self.install({client: data}, installed_packages, False, now)

    def ondemand(self, clients):
        """Trigger ondemand event on clients.

        :param dict clients: client dict with name
        """

        logging.debug("app.py:ondemand event on %s" % list(clients))

        self.__backend.ondemand(list(clients))

    def reboot(self, clients):
        """Reboot clients.

        :param dict clients: client dict with name
        """

        logging.debug("app.py:rebooting %s" % list(clients))

        self.__backend.reboot(list(clients))

    def shutdown(self, targets):
        """Shutdown clients.

        :param dict clients: client dict with name
        """

        logging.debug("shutting down %s" % list(targets))

        self.__backend.shutdown(list(targets))

    def updates_(self, ssh_cmd):
        """Listet für das Depot verfügbare Paket-Updates auf"""

        proc = subprocess.Popen(f"ssh opsi-adminuser opsi-package-updater list --updatable-packages",
                shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        if proc.returncode < 0:
            print("Fehler:")
            print(stderr)
        else:
            # stdout is a byte stream
            return stdout.decode()

    def install_updates_on_depot_(self, ssh_cmd):
        """Installiert für das Depot verfügbare Paket-Updates"""

        proc = subprocess.Popen(f"ssh opsi-adminuser opsi-package-updater update",
                shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        if proc.returncode < 0:
            print("Fehler:")
            print(stderr)
        else:
            # stdout is a byte stream
            return stdout.decode()

    def unpackaged(self, client, json):
        """Get a list of installed, but unpackaged software on a client.

        This is a somewhat manual approach as it requires a custom list of
        software names ('_mapping').  Helpful for migration into opsi.

        :param str client: client id
        :param dict json: client json data
        :return: dictionary with name:list of unpackaged software
        """

        logging.debug("app.Opus.unpackaged: %s" % client)

        _, unpackaged_sw = self.list_migratable(client, json)

        return unpackaged_sw

    def set_installed(self, clients, packages, installed=True):
        """Set packages to 'installed' state on given clients.

        :param list clients: client ids
        :param list packages: package ids
        """

        logging.debug("app.Opus.set_installed (%s)" % installed)

        # first check if all packages are available on depot
        # raise an exception if a package does not exist, otherwise collect them
        # in a dictionary for later
        logging.debug("app.Opus.set_installed: checking available packages")
        for p in packages:
            if not p in self.products():
                raise NoSuchProduct(p)

        for client in clients:
            for p in packages:
                if installed:
                    self.__backend.setInstalled(client, p)
                else:
                    self.__backend.setUninstalled(client, p)

    def set_setup(self, clients, packages):
        """Set packages to 'setup' state on given clients.

        :param list clients: client ids
        :param list packages: package ids
        """

        logging.debug("app.Opus.set_setup")

        # first check if all packages are available on depot
        # raise an exception if a package does not exist, otherwise collect them
        # in a dictionary for later
        logging.debug("app.Opus.set_setup: checking available packages")
        for p in packages:
            if not p in self.products():
                raise NoSuchProduct(p)

        for client in clients:
            for p in packages:
                self.__backend.setSetup(client, p)


    def os_version(self, client):
        """Get OS version info for a given client.

        This requires a software audit beforehand.

        :param client: client id
        :return: string with version info or None
        """

        logging.debug("app.Opus.os_version: client %s" % client)

        software = self.__backend.softwareOnClient(client)

        for s in software:
            if s.startswith("Windows 1"):
                return s
        return None


    def list_migratable(self, client, json):
        """Get a list of installed software with available opsi packages for
        a given client.

        :param string client: client id
        :param dict json: client json data
        :return: not installed available packages in depot, missing packages
        """

        logging.debug("app.Opus.list_migratable: client %s" % client)

        available_packages = set()
        missing_packages = []

        mapping = _build_mapping()
        # get list of software generated by swaudit
        software = self.__backend.softwareOnClient(client)

        # see if the found software names are in our mapping
        found = False
        for s in software:
            name = s.lower()
            for key in mapping.keys():
                if name.startswith(key):
                    # prevent multiple entries for names that belong to one package, e.g.
                    # sophos antivir, sophos endpoint etc.
                    if not mapping[key] in available_packages:
                        available_packages.update([mapping[key]])
                    found = True
                    break

            if not found:
                missing_packages.append(s)
            else:
                found = False

        # remove already installed opsi packages from list 
        installed_packages = set()
        for product_triple in json["products"]:
            installed_packages.update([product_triple[0]])

        # remove already installed packages from available
        return available_packages - installed_packages, missing_packages

    def migrate(self, clients, masked_pkgs):
        """Mark software that has an opsi package as 'installed'.

        :param list clients: list of client ids
        """

        for client, json in clients.items():
            packages, _ = self.list_migratable(client, json)
            #self.set_installed([client], packages, True)
            for package in packages:
                # Pakete maskieren
                if package in masked_pkgs:
                    continue
                self.__backend.install(client, package)

# custom made list of packages
_mapping = """
2;activate-win                   1.0
2;config-win10                   4.1.2
3;config-win81-desktop           4.0.5
4;config-win-base                4.0.1
5;dfn_adobe-air                  32.0.0.125
8;dfn_ant-renamer                2.12.0
9;dfn_atlas.ti                   8.3.20
10;dfn_camstudio                  2.7.316
11;dfn_carotdav                   1.16.1
12;dfn_ccleaner                   5.91
13;dfn_cdburnerxp                 4.5.8.7128
14;dfn_cdrtfe                     1.5.9
15;dfn_chromas-free               2.6.6
16;dfn_citavi                     6.14.0.0
17;dfn_claws-mail                 4.0.0.1
18;dfn_cmake                      3.25.0
19;dfn_context                    1.2.0
20;dfn_crystaldiskinfo            8.17.11
21;dfn_darktable                  4.0.1
22;dfn_dotnet                     4.8.3761.0
23;dfn_doublecmd                  1.0.6
24;dfn_dwgtrueview                2019.23.0.46.0
25;dfn_endnote                    20.2.1
26;dfn_evernote                   6.25.1.9091
27;dfn_f4transkript               7.0.6
28;dfn_filezilla                  3.62.2
29;dfn_flashplayer-esr            32.0.0.465
30;dfn_freemind                   1.0.1
31;dfn_gajim-messenger            1.5.4
32;dfn_ganttproject               3.2.3230
33;dfn_gitscm                     2.38.0
34;dfn_glasswire                  3.0.476.0
35;dfn_greenshot                  1.2.10.6
gsview;dfn_gsview
37;dfn_handbrake                  1.5.1
38;dfn_joplin-notice              2.8.8
39;dfn_jre                        8.0.201.09
40;dfn_jxplorer                   3.3.1.2
41;dfn_keepass                    2.52
42;dfn_kitty                      0.76.0.8
43;dfn_krita                      5.1.3
44;dfn_libreoffice-still          7.1.8
45;dfn_licor-fileviewer           1.0.4
46;dfn_linphone                   4.4.1
47;dfn_luminance                  2.6.0
48;dfn_lupas-rename               5.0
49;dfn_matlab                     2017.2
50;dfn_maxqda                     20.4.2
51;dfn_maxqda-reader              18.1.0
52;dfn_mega                       10.0.5
53;dfn_mendeley-desktop           1.19.8
54;dfn_miktex-basic               22.3
55;dfn_mirandang-messenger        0.95.13.1
56;dfn_mobaxterm                  22.3
57;dfn_mremoteng                  1.76.20.24615
mu;dfn_mu-editor
59;dfn_myphoneexplorer            1.9.0
60;dfn_nagstamon                  3.10.1
61;dfn_nextcloud                  3.4.1
62;dfn_onlyoffice-desktopeditor   7.2.1
63;dfn_open365                    1.0.0.153934
65;dfn_openvpn                    2.4.9
66;dfn_opsipackagebuilderng       8.4.2
67;dfn_owncloud                   2.7.6.3261
68;dfn_pgadmin                    4.5.1
df69;n_pidgin                     2.14.10
70;dfn_projectlibre               1.9.3.0
71;dfn_protege-desktop            5.5.0
72;dfn_pspp                       20181109
73;dfn_publishorperish            8.4.4041
74;dfn_putty                      0.75
75;dfn_qnap-qfinderpro            7.8.2.0928
76;dfn_r                          4.2.2
77;dfn_rstudio                    2022.07.2.576
78;dfn_shotcut                    221125
79;dfn_signal-messenger           6.0.1
80;dfn_skype                      8.92.0.204
81;dfn_spark                      3.0.0
82;dfn_srware-iron                84.0.4300.0
83;dfn_sumatra                    3.3.3
84;dfn_syncbackfree               10.2.49.0
85;dfn_texniccenter               2.02
86;dfn_therakles                  3.2.1
87;dfn_thinkpad-systemupdate      5.07.0131
88;dfn_tortoisegit                2.13.0.0
89;dfn_tortoisesvn                1.14.3.29387
90;dfn_vidiot                     0.3.37
91;dfn_virtualbox                 7.0.4.154605
92;dfn_vsdc-videoeditor           7.2
93;dfn_win32diskimager            1.0.0
94;dfn_winhttrack                 3.49.2
95;dfn_winscp                     5.17.6
96;dfn_wireshark                  4.0.2
x2go client for windows;dfn_x2go-client
98;dfn_zotero                     6.0.18
99;jedit                          5.5.0
100;networklocation                4.2.0.0
7-zip;o4i_7zip
7-zip 22.01 (x64 edition);o4i_7zip
102;o4i_acrobat-reader-classic     20.005.30418
103;o4i_audacity                   3.2.2
104;o4i_avidemux                   2.8.2
105;o4i_blender-lts                3.3.2
106;o4i_caldavsynchronizer         4.3.0
107;o4i_cdburnerxp                 4.5.8.7128
108;o4i_cdrtfe                     1.5.9
google chrome;o4i_chrome
110;o4i_claws-mail                 4.0.0.1
111;o4i_clickmonitorddc            7.0
112;o4i_corretto                   11.0.17.8.1
dell command | update for windows universal;o4i_dell-command-update
114;o4i_dell-display-manager       1.52.2068
115;o4i_dell-displaymanager        1.53.2109
116;o4i_doublecmd                  1.0.9
element (installed by opsi);o4i_element-messenger
118;o4i_faststone-imageviewer      7.6
119;o4i_filius                     1.13.1
mozilla firefox esr;o4i_firefox-esr
mozilla firefox esr (x64 de);o4i_firefox-esr
mozilla firefox (x86 de);o4i_firefox-esr
121;o4i_freecad                    0.20.1
gpl ghostscript;o4i_ghostscript
123;o4i_gimp                       2.10.32
greenshot;o4i_greenshot
125;o4i_hdfview                    3.2.0
126;o4i_hugin                      2021.0
127;o4i_inkscape                   1.2.1
irfanview;o4i_irfanview
129;o4i_iron                       106.0.5400.0
jabref;o4i_jabref
131;o4i_kaspersky                  12
keepassxc;o4i_keepassxc
133;o4i_kitty                      0.76.0.14
134;o4i_klayout                    0.26.7
libreoffice;o4i_libreoffice-still
136;o4i_luminance                  2.6.0
137;o4i_matlab                     2020.2
138;o4i_meshlab                    2020.06
139;o4i_miktex-basic               22.10
140;o4i_mobaxterm                  22.3
microsoft edge;o4i_msedge-browser
142;o4i_nextcloud                  3.6.4
143;o4i_notepadpp                  8.4.7
144;o4i_open-shell                 4.4.170
145;o4i_openvpn                    2.5.8.603
146;o4i_owncloud                   3.0.0.9215
sciebo;o4i_owncloud
147;o4i_paint-dot-net              4.3.12
148;o4i_pdf24creator               11.8.0
149;o4i_pgadmin                    4.6.17
150;o4i_pidgin                     2.14.7
putty;o4i_putty
152;o4i_qgis-ltr                   3.22.12
153;o4i_r                          4.2.2
154;o4i_rocketchat                 3.8.9
155;o4i_rstudio                    2022.07.2.576
156;o4i_scribus                    1.4.8
157;o4i_signal-messenger           5.0.0
sumatrapdf;o4i_sumatra
159;o4i_surun                      1.2.1.4
160;o4i_teamviewer                 15.33.7
texstudio - texstudio is a fully featured latex editor;o4i_texstudio
162;o4i_thinkpad-systemupdate      5.07.0140
mozilla thunderbird (x64 de);o4i_thunderbird
mozilla thunderbird (x86 de);o4i_thunderbird
164;o4i_ueli                       8.22.0
165;o4i_vivaldi                    5.6.2867.40
vlc media player;o4i_vlc
windirstat;o4i_windirstat
168;o4i_winibw                     3.7.1.13
winmerge;o4i_winmerge
winscp;o4i_winscp
zoom;o4i_zoom-client
acrobat pro dc;uwiw-adobe
adobe acrobat dc;uwiw-adobe
adobe acrobat;uwiw-adobe
cisco anyconnect secure mobility client;uwiw-anyconnect
175;uwiw-atlas.ti                  8.3.20
autohotkey;uwiw-autohotkey
local administrator password solution;uwiw-laps
microsoft office professional plus 2019 - de-de;uwiw-office
microsoft office professional plus 2019 - en-us;uwiw-office
sophos;uwiw-sophos
veracrypt;uwiw-veracrypt
office;uwiw-office
opsi-client-agent;opsi-client-agent
"""

def _build_mapping():

    mapping = {}
    for line in _mapping.split("\n"):
        line = line.strip()
        if not len(line) > 0:
            continue
        key, value = line.split(";")
        mapping[key] = value

    return mapping

def gt(a, b):
    """Compare two packages.

    a.name==b.name and a.version>b.version and
        a.package_version==b.package_version

    → a>b

    :param dict a: depot package entry (dictionary)
    :param tuple b: product triple
    """

    logging.debug("app.gt: %s > %s?" % (a["id"], b[0]))
    # only the same product can be compared
    if not a["id"] == b[0]:
        logging.debug("app.gt: %s > %s = False" % (a["id"], b[0]))
        return False

    # check for None -> unknown installationStatus = failure
    if a["productVersion"]!=None and b[1]==None:
        logging.debug("app.gt: a has installationStatus, b not -> a>b=True")
        return True
    elif a["productVersion"]==None and b[1]==None:
        logging.debug("app.gt: both have no installationStatus -> a>b=False")
        return False
    else:
        logging.debug("app.gt: both are installed")


    # check if there are only digits; needed for win 21h2 etc.
    # do not forget to remove dots
    version_a = a.get("productVersion", "")
    version_b = b[1] or ""

    if not version_a.replace(".", "").isdecimal() or \
            not version_b.replace(".", "").isdecimal():

        logging.debug("app.gt: version a > version b = %s" % version_a > version_b)
        return version_a > version_b

    parts1 = a["productVersion"].split(".")
    parts2 = b[1].split(".")
    index = 0

    logging.debug("app.gt: looking in details: %s > %s" % (parts1, parts2))
    a_equal_b = False
    while True:
        # check if there is a part that can be compared
        if len(parts1)>index and len(parts2)>index:
            # ok, which one is greater?
            if int(parts1[index])>int(parts2[index]):
                return True
            elif int(parts1[index])==int(parts2[index]):
                logging.debug("app.gt: same version at pos %d: %s==%s" % (index, parts1[index], parts2[index]))
                index +=1
                a_equal_b = True
                continue
            else:
                return False
        else:
            if a_equal_b:
                logging.debug("app.gt: same version -> return False")
                return False

            # either both versions string have the same number of parts
            # (then rely on the package version) ...
            if len(parts1)==len(parts2):
                return int(a["packageVersion"])>int(b[2])

            # ... or the first parts are equal, but one string is longer than
            # the other
            # the longer one wins
            elif len(parts1)>len(parts2):
                return True
            else:
                return False

