"""
Backend
-------

Das `backend`-Modul definiert mit der Klasse `Backend` die grundlegenden
Methoden für die Anwendung, ist dabei aber eine vergleichsweise dünne
Abstraktionsschicht, die sich die Funktionalität von :py:mod:`opus.server.Server`
zunutze macht.
"""

import logging

from opus import server
from opus.exceptions import *


class Backend:
    """Die Backend-Klasse hält die Verbindungsdaten zum Server vor und stellt
    die grundlegenden Methoden für die Arbeit mit opsi-Clients bereit.

    Anwendungsbeispiel:

    .. code-block:: python

        backend = Backend(url, username, password)

        # known clients
        clients = backend.clients
        # installed products on client 'x'
        products_on_client_x = client["x"].products
    """

    def __init__(self, url, username, password):
        self.__server = server.Server(url, username, password)
        self.__clients = {}
        self.__clientGroups = {}
        # global products list
        self.__products = {}
        self.__productGroups = {}

        logging.debug("backend.Opus.__init__: built Backend instance")


    @property
    def products(self):
        """get available products

        :return: list of products (dictionaries)
        """

        logging.debug("backend.Opus.products: getting products")

        if len(self.__products.keys()) == 0:
            self.__refresh()
        return self.__products

    @property
    def clients(self):
        """get clients

        :return: dictionary with client id:data; data is a dictionary
            with `products`:[(productId, productVersion, packageVersion,
            installationStatus, lastAction, actionResult)]
        """

        logging.debug("backend.Opus.clients: getting clients")

        if len(self.__clients) == 0:
            self.__refresh()

        return self.__clients

    def __refresh(self):
        """updating internal cache for clients and products"""

        logging.debug("backend.Opus.__refresh: refreshing clients and products")

        for product in self.__server.products():
            self.__products[ product["id"] ] = product
            # 'product' is the JSON answer; for visualcpp-2017 is this:
            #
            # {'name': 'Microsoft Visual C++ 2017', 'licenseRequired': False,
            # 'setupScript': 'setup3264.opsiscript', 'uninstallScript': '',
            # 'updateScript': '', 'alwaysScript': '', 'onceScript': '',
            # 'customScript': '', 'userLoginScript': '', 'priority': 0,
            # 'description': 'Microsoft Visual C++ 2017 redistributable Installer 32- und 64-bit',
            # 'advice': '', 'changelog': '', 'productClassIds': [], 'windowsSoftwareIds': [],
            # 'id': 'visualcpp-2017', 'productVersion': '14.12.25810.0', 'packageVersion': '1',
            # 'type': 'LocalbootProduct', 'ident': 'visualcpp-2017;14.12.25810.0;1'}

        for client in self.__server.clients():
            self.__clients[ client["id"] ] = client

        # we could only need to get the list with clients and their products and
        # not a separate client list, because every client has at least
        # the opsi-client-agent and the opsi-script product installed. but
        # we miss the last ip, last seen etc.!
        # use a list here and therefore ONE jsonprc call to reduce server load
        products_on_clients = self.__server.productsOnClient(list(self.__clients))
        logging.debug("found %d products on client(s)" % len(products_on_clients))

        # iterate over found products and assign them as a tuple to the clientId entry
        for entry in products_on_clients:
            client_id = entry["clientId"]
            if not "products" in self.__clients[client_id]:
                self.__clients[client_id]["products"] = []

            # only successfully installed products are relevant
            if entry["installationStatus"]=="installed":
                product = ( entry["productId"], entry["productVersion"], \
                    entry["packageVersion"], entry["installationStatus"], \
                    entry["lastAction"], entry["actionResult"] )
                self.__clients[client_id]["products"].append(product)

        # check for clients that are NOT in the list because they do not have any products installed
        # this can happen sometimes for unknown reasons
        for id_, client in self.clients.items():
            if not "products" in client:
                client["products"] = []

    def clientGroups(self, group_id=""):
        """get client groups

        :param str group_id: group id
        :return: dictionary with (group) id and description
        """

        logging.debug("backend.Opus.clientGroups: building client groups")

        if not len(self.__clientGroups):
            groups = self.__server.clientGroups(group_id)
            for group in groups:
                self.__clientGroups[group["id"]] = group["description"]

        return self.__clientGroups

    def clientGroupMembers(self, group_id):
        """get members of the given client group

        :param str group_id: group id
        :return: list with client ids
        """

        logging.debug("backend.Opus.clientGroupMembers: getting members of group %s" % group_id)

        if not len(self.__clients):
            self.__refresh()

        return self.__server.clientGroupMembers(group_id)

    def addClientsToGroup(self, client_ids, group_id):
        """add client(s) to the given group

        :param list client_ids: client ids
        :param str group_id: group id
        """

        logging.debug("backend.Opus.addClientsToGroup: adding %s to group %s" % (client_ids, group_id))

        for client_id in client_ids:
            self.__server.addClientToGroup(client_id, group_id)

    def productGroups(self):
        """get list of known product groups

        :return: dictionary with product group ids and descriptions
        """

        logging.debug("backend.Opus.productGroups: getting product group ids")

        if not len(self.__productGroups):
            groups = self.__server.productGroups()
            for group in groups:
                self.__productGroups[group["id"]] = group["description"]

        return self.__productGroups

    def productGroupMembers(self, group_id):
        """get list of members of the given product group

        :param str group_id: group id
        :return: list of group member ids
        """

        logging.debug("backend.Opus.productGroupMembers: getting members of group %s" % group_id)

        member_ids = [member["objectId"] for member in self.__server.productGroupMembers(group_id)]
        return member_ids

    def install(self, client_id, product_id):
        """install product on client

        :param str client_id: client id
        :param str product_id: product id
        """

        logging.debug("backend.Opus.install: installing %s on %s" % (product_id, client_id))

        self.__server.installProductOnClient(client_id, product_id)

    def uninstall(self, client_id, product_id):
        """remove product from client

        :param str client_id: client id
        :param str product_id: product id
        """

        logging.debug("backend.Opus.uninstall: uninstalling %s on %s" % (product_id, client_id))

        self.__server.uninstallProductOnClient(client_id, product_id)

    def setSetup(self, client_id, product_id):
        """set product on client to setup

        :param str client_id: client id
        :param str product_id: product id
        """

        logging.debug("set %s for %s to 'setup" % (product_id, client_id))

        self.__server.setProductOnClientStatus(client_id, product_id, status="setup")

    def setInstalled(self, client_id, product_id):
        """install product on client

        :param str client_id: client id
        :param str product_id: product id
        """

        logging.debug("set %s for %s to 'installed" % (product_id, client_id))

        self.__server.setProductOnClientStatus(client_id, product_id, status="installed")

    def setUninstalled(self, client_id, product_id):
        """set product status to 'uninstalled on client(s)install product on
        client

        :param str client_id: client id
        :param str product_id: product id
        """

        logging.debug("backend.Opus.setUninstalled: set %s for %s to 'uninstalled" % (product_id, client_id))

        self.__server.setProductOnClientStatus(client_id, product_id, status="not_installed")

    def ondemand(self, client_ids):
        """trigger ondemand event on client(s)

        :param list client_ids: list with client ids
        """

        logging.debug("backendOpus.ondemand: triggering 'ondemand' on %s" % client_ids)

        self.__server.fireEvent(client_ids, event="on_demand")

    def reboot(self, client_ids):
        """reboot client(s)

        :param list client_ids: list with client ids
        """

        logging.debug("backend.Opus.reboot: rebooting %s" % client_ids)

        self.__server.reboot(client_ids)

    def shutdown(self, client_ids):
        """shutdown clienti(s)

        :param list client_ids: list with client ids
        """

        logging.debug("backend.Opus.shutdown: shutting down %s" % client_ids)

        self.__server.shutdown(client_ids)

    def softwareOnClient(self, client):
        """get list of installed software (not only by opsi, but all software)
        on a client.

        Warning: This requires a run `swaudit` beforehand!

        :param str client: client id
        :return: list with dicts
        """

        logging.debug("backend.Opus.refreshSoftwareList for %s" % client)

        # extract names
        found = self.__server.softwareOnClient(client)
        return  [ x["ident"].split(";")[0] for x in found]

