"""

Read configuration file

Usage::

    config_obj = get_config(path_to_file)

"""

import configparser
import logging
import os
import os.path
import sys


class Config:
    """
    Konfigurations

    ~/.config/opus/config

    Beispiel::

        [DEFAULT]
        depot = mein_depot
        masked = 

        [mein_depot]
        url = https://opsi.server:4447/rpc
        username = adminuser
        password = geheim123

        [uib]
        url = https://opsi.uib:4447/rpc
        username = User
        password = Passwort

    """
    def __init__(self):
        self.__depots = {}
        self.default = {
            "url": "",
            "username": "",
            "password": "",
            "masked": [],
        }
        logging.debug("cfg: built instance")

    @property
    def depots(self):
        return self.__depots.keys()

    @property
    def username(self):
        return self.default["username"]

    @property
    def password(self):
        return self.default["password"]

    @property
    def url(self):
        return self.default["url"]

    @property
    def masked(self):
        return self.default["masked"]

    @property
    def ssh_cmd(self):
        return self.default["ssh_cmd"]

    def parse(self, path=None):
        default_config = {
            "url": "",
            "username": "",
            "password": "",
            "masked": [],
        }
        logging.debug("cfg.parse")

        # default configuration file is $HOME/.config/opus/config
        if not path:
            home_dir = os.environ["HOME"]
            path = os.path.join(home_dir, ".config", "opus", "config")
            logging.debug("cfg.parse: using default path %s" % path)

        parser = configparser.ConfigParser()
        try:
            config_file = open(path, "r")
        except FileNotFoundError as err:
            # use defaults instead
            print('config file not found, using defaults', file=sys.stderr)
        else:
            parser.read_file(config_file)
            config_file.close()

            # read sections with depot infos
            for section in parser.sections():
                cfg = {}
                for key, value in parser[section].items():
                    if key == "masked":
                        cfg[key] = value.split()

                    else:
                        cfg[key] = value
                self.__depots[section] = cfg

            logging.debug("cfg.parse: using default path %s" % path)
            try:
                default_section = parser['DEFAULT']
            except KeyError as err:
                print('Missing [DEFAULT] section in config file.  Using first section as default, if available; defaults otherwise.')
                if len(self.__depots):
                    self.default = self.__depots[self.__depots.keys()[0]]
            except configparser.MissingSectionHeaderError as err:
                print('Error in config file', path, file=sys.stderr)
                print('Using defaults.')
            else:
                depot = default_section['depot']
                self.default = self.__depots[depot]


