"""
Server
------

Das `server`-Modul kapselt die Kommunikation mit opsi-Servern.

Die Kommunikation über das JSON-RPC-Backend übernimmt die Klasse
:class:`opus.server.Server`, sie agiert quasi als DB-Schicht.
"""

import logging
import urllib3
# disable warnings for opsi self certificate
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import ConnectionError, RequestException

from .exceptions import *


class Server:
    """Server-Verbindung.

    Eine `Server`-Instanz hält die Zugangsdaten für einen JSONRPC-Aufruf,
    öffnet allerdings keine Verbindung.  Dies erfolgt in den jeweiligen
    Methoden, die eine Payload vorbereiten und an `__rpc_request` übergeben.

    ::

        server = Server(url, username, password)
        client_ids = server.clients()

    """

    def __init__(self, url: str, username: str, password: str):
        """create instance

        :param str url: opsi server url
        :param str username: user name for login
        :param str password: password for login
        """

        self.client = None
        self.backend = None
        self.id_ = 0
        self.url = url + "/rpc"
        self.username = username
        self.password = password

    def __str__(self):
        return self.url

    def __build_session(self):
        """build a session object with given credentials"""

        session = requests.Session()
        session.trust_env = False
        session.auth = (self.username, self.password)

        logging.debug("server.Server.__session: created session")
        return session

    def __rpc_request(self, payload: dict):
        """make a RPC request with the given payload

        :param dict payload: JSON dict
        :return: JSONRPC result
        """

        logging.debug("server.Server.__rpc_request: starting rpc request with %s" % payload)

        self.id_ += 1
        payload["id"] = self.id_

        try:
            # create new session
            with self.__build_session() as session:
                # use verify=False for self-signed certificates
                result = session.post(self.url, json=payload, verify=False)
                # close session to reduce server load?
                result.raise_for_status()
        except requests.exceptions.RequestException as err:
            raise err

        if result.json()["error"]:
            logging.error(result.json()["error"]["message"])
            raise RPCError("%s: %s: %s" % (payload,
                result.json()["error"]["class"],
                result.json()["error"]["message"]))
        else:
            return result.json()["result"]

    def clients(self):
        """return a list with all client ids

        :return: list with opsi client infos
        """

        logging.debug("server.Server.clients")

        payload =  {
            "method":"host_getObjects",
            "params": [
                [],                     # attributes
                {"type": "OpsiClient"}  # filter
            ]
        }

        clients = self.__rpc_request(payload)

        return clients

    def clientGroups(self, group_id=""):
        """return a list of client groups

        :param str group_id: group id
        :return: list of HostGroup instances

        Relevant attributes/methods:

        * `id`: id of the group; index 0 is the main `clientdirectory`
        * `description`: group description
        """

        logging.debug("server.Server.clientGroups")

        payload =  { "method":"group_getObjects" }
        if len(group_id):
            payload["params"] = [
                [],
                {
                    "type": "HostGroup",
                    "parentGroupId": group_id,
                }]
        else:
            payload["params"] = [
                [],
                {
                    "type": "HostGroup",
                }]

        groups = self.__rpc_request(payload)
        return groups

    def clientGroupMembers(self, group_id):
        """return a list with clients in a group

        :param str group_id: group id
        :return: list of ObjectToGroup instances (these a Client instances)

        Relevant attributes:

        * `objectId`: id of the Client instance
        """

        logging.debug("server.Server.clientGroupMembers")

        payload =  { "method": "objectToGroup_getObjects" }
        if len(group_id):
            payload["params"] = [
                [],
                {
                    "groupType": "HostGroup",
                    "groupId": group_id,
                }]
        else:
            payload["params"] = [
                [],
                {
                    "groupType": "HostGroup",
                }]

        members = self.__rpc_request(payload)
        return [ member["objectId"] for member in members ]

    def addClientToGroup(self, client_id, group_id):
        """add a client to a group

        :param str client_id: client id
        :param str group_id: group id
        :return: JSONRPC result
        """

        logging.debug("server.Server.addClientToGroup")

        payload =  { "method": "objectToGroup_create" }
        payload["params"] = [
                "HostGroup",
                group_id,
                client_id
        ]

        result = self.__rpc_request(payload)
        return result

    def products(self):
        """return a list of all available products

        :return: list of LocalbootProduct products
        """

        logging.debug("server.Server.products")

        payload =  {
            "method":"product_getObjects",
            "params": [
                [],
                {"type": "LocalbootProduct"}
            ]
        }

        products = self.__rpc_request(payload)
        return products

    def productGroups(self):
        """return a list of known product groups

        :return: list with ProductGroup instances
        """

        logging.debug("server.Server.productGroups")

        payload =  {
            "method":"group_getObjects",
            "params": [
                [],
                {"type": "ProductGroup"}
            ]
        }

        product_groups = self.__rpc_request(payload)
        return product_groups

    def productGroupMembers(self, group_id=""):
        """return a list with products of a group

        :param str group_id: product group name
        :return: list with product ids
        """

        logging.debug("server.Server.productGroupMembers")

        payload =  { "method":"objectToGroup_getObjects" }
        if len(group_id):
            payload["params"] = [
                [],
                {
                    "groupType": "ProductGroup",
                    "groupId": group_id,
                }]
        else:
            payload["params"] = [
                [],
                {
                    "groupType": "ProductGroup",
                }]

        groups = self.__rpc_request(payload)
        return groups

    def productsOnClient(self, client_id=""):
        """return a list with all products installed on client(s)

        :param str client_id: client id; leave empty for all clients
        :return: list with ProductOnClient instances

        Helpful attributes:

        * `productId`: product id, needed for `installProductOnClient`
        * `version`: product version with packet version
        * `productVersion`: product version alone
        """

        logging.debug("server.Server.productsOnClient")

        payload =  { "method":"productOnClient_getObjects" }
        if len(client_id):
            payload["params"] = [
                [],
                {
                    "clientId": client_id
                }]
        else:
            payload["params"] = [ [], { }]

        products = self.__rpc_request(payload)
        return products

    def productsOnDepot(self):
        """list with available products

        :return: list with available products in depot
        """

        logging.debug("server.Server.productsOnDepot")

        payload =  {
            "method":"productOnDepot_getObjects",
            "params": [ [], { }] }

        products = self.__rpc_request(payload)
        return products


    def installProductOnClient(self, client_id, product_id):
        """install a product on a client

        :param str client_id: client id
        :param str product_id: product id
        :return: JSONRPC result
        """

        logging.debug("server.Server.installProductsOnClient")

        payload =  {
            "method":"productOnClient_create",
            "params": [
                    product_id,
                    "LocalbootProduct",
                    client_id,
                    "",
                    "setup",
            ]
        }

        result = self.__rpc_request(payload)
        return result

    def softwareOnClient(self, client_id):
        """get all installed products on given client

        :param str client_id: client id
        :return: JSONRPC result
        """

        logging.debug("server.Server.softwareOnClient for %s" % client_id)

        payload =  {
            "method": "auditSoftwareOnClient_getObjects",
            "params": [
                [],
                { "clientId": client_id }]
        }

        result = self.__rpc_request(payload)
        return result

    def setProductOnClientStatus(self, client_id, product_id, status="installed"):
        """set the product status on a client; default: installed

        :param str client_id: client id
        :param str product_id: product id
        :param str status: status to set
        :return: JSONRPC result
        """

        logging.debug("server.Server.setProductOnClientStatus: %s" % status)

        payload =  {
            "method":"productOnClient_create",
            "params": [
                    product_id,
                    "LocalbootProduct",
                    client_id,
                    status
            ]
        }

        result = self.__rpc_request(payload)
        return result

    def uninstallProductOnClient(self, client_id, product_id):
        """remove a product from a client

        :param str client_id: client id
        :param str product_id: product id
        :return: JSONRPC result
        """

        logging.debug("server.Server.uninstallProductOnClient: %s: %s" % (client_id, product_id))

        payload =  {
            "method":"productOnClient_create",
            "params": [
                    product_id,
                    "LocalbootProduct",
                    client_id,
                    "",
                    "uninstall",
            ]
        }


        result = self.__rpc_request(payload)
        return result

    def fireEvent(self, client_ids, event="on_demand"):
        """fire the given event for given clients

        :param list client_ids: list with client ids
        :param str event: event (default: on_demand); values are:  panic,
            software_on_demand, sync_completed, on_shutdown, silent_install,
            gui_startup
        :return: JSONRPC result
        """

        logging.debug("server.Server.fireEvent: %s: %s" % (client_ids, event))

        payload =  {
            "method":"hostControlSafe_fireEvent",
            "params": [
                "on_demand",
                client_ids
                ]
        }

        result = self.__rpc_request(payload)
        return result

    def reboot(self, client_ids):
        """reboot clients

        :param list client_ids: list with client ids
        :return: JSONRPC result
        """

        logging.debug("server.Server.reboot: %s" % client_ids)

        payload =  {
            "method":"hostControlSafe_reboot",
            "params": [ client_ids]
        }

        result = self.__rpc_request(payload)
        return result

    def shutdown(self, client_ids):
        """shutdown clients

        :param client_ids: list with client ids
        :return: JSONRPC result
        """

        logging.debug("server.Server.shutdown: %s" % client_ids)

        payload = {
            "method":"hostControlSafe_shutdown",
            "params": [ client_ids ]
        }

        result = self.__rpc_request(payload)
        return result

