"""
Targetting of Clients
---------------------

Possible targets of an operation can be specified

* directly by id (single client)
* via globbing (e.g. `*`)
* regular expression: `E@uwiw-.*`
* named group of clients: `N@uwiw-group`
* list of clients: `L@a,b`


Usage
-----

.. code-block:: python

	target_str = "*" # or L@uwiw-1,utech-2 etc.
	targets = backend.evaluate_targets(target_str)

"""

import fnmatch
import logging
import re


def evaluate(target_string, clients, app):
    """

    :param: target_string - targeting string
    :param: clients - dictionary with client id:JSON
    :param: app - Opus instance, needed for calling `groupMembers`
    
    :return: list of client ids
    """

    if len(target_string)==0:
        return []
    logging.debug("evaluating '%s' with clients %s" % (target_string, clients))
    if target_string.startswith("E@"):
        return _regexp(target_string[2:], clients)
    elif target_string.startswith("N@"):
        return _group(target_string[2:], app, clients)
    elif target_string.startswith("L@"):
        return _list(target_string[2:], clients)
    else:
        return _glob(target_string, clients)


def _glob(target_string, clients):
    """Find target(s) with globbing; uses ID extact id matching as fallback

    :param: target_string - globbing string
    :param: clients - dictionary

    :return: dictionary
    """

    logging.debug("using globbing for target search")
    regexp = fnmatch.translate(target_string)
    reobj = re.compile(regexp)

    targets = {}
    for client_id, client in clients.items():
        if reobj.match(client_id):
            targets[client_id] = client

    # ok, no globbing, perhaps ID?
    logging.debug("using exact match as fallback for globbing")
    if len(targets)==0:
        for client_id, client in clients.items():
            name = client_id.split(".")[0]
            if target_string == name:
                targets[client_id] = client
                break

    return targets


def _regexp(regexp, clients):
    """find target(s) with regular expression

    :param: regexp - regular expression
    :param: clients - dictionary

    :return: dictionary
    """
    logging.debug("using regular expressions for target search")
    reobj = re.compile(regexp)

    targets = {}
    for client_id, client in clients.items():
        if reobj.match(client_id):
            targets[client_id] = client

    return targets


def _group(group_ids, app, clients):
    """find all clients belonging to the given group(s)

    :param: group_ids - comma-separated list of group ids
    :param: app - Opus instance
    
    :return: list  with clients belonging to the given group(s)
    """
    logging.debug("targeting._group: using groups ids for target search")
    member_clients = {}
    for gid in group_ids.split(","):
        # list with ids
        members = app.clientGroupMembers(gid)
        for member in members:
            member_clients[member] = clients[member]

    return member_clients


def _list(client_ids, clients):
    """return all clients in list

    :param: client_ids - list of clients
    :param: clients - dictionary

    :return: dictionary
    """

    logging.debug("using id list for target search")
    targets = {}
    for client_id in client_ids.split(","):
        targets[client_id] = clients[client_id] 

    return targets

