"""
UI
--

Das `ui`-Modul stellt eine Konsolen-Schnittstelle zu opus bereit.  Es nutzt
das `click-Paket <https://click.palletsprojects.com/>`_, um die auf der
Befehlszeile übergebenen Parameter auszuwerten.
"""

import logging
#logging.basicConfig(format='%(asctime)s %(filename)s/%(module)s.%(funcname)s %(lineno)d: %(message)s')
# wird von click_logging nicht berücksichtigt
import sys

import click
import click_logging
from requests.exceptions import ConnectionError

from opus import __version__, __prog_name__
from opus import app, cfg
from opus.exceptions import *

# set logging helper vor click
logger = logging.getLogger()
click_logging.basic_config(logger)

csv_output = False


def get_login():
    """Get credentials for login.

    :return: (username, password)
    """

    from getpass import getpass
    from prompt_toolkit import prompt

    username = prompt("Username: ")
    password = getpass()

    return username, password


def get_credentials(conf):
    """Get credentials either from configuration file or via dialog.

    :param str conf: path to configuration file
    :return: triple (url, username, password)
    """

    logger.debug("ui.get_credentials: gettings credentials")
    config = cfg.Config()
    config.parse(conf)

    url = config.url
    masked_pkgs = config.masked
    username = config.username
    password = config.password
    #ssh_cmd = config.ssh_cmd
    if username=="" or password=="":
        username, password = get_login()

    return url, username, password, masked_pkgs


def print_information(url, username):
    """Print login information."""

    click.echo(f"{url}, {username}")


def install(o, clients, products, force=False, now=False):
    """Install products on clients.

    :param Opus o: Opus instance
    :param dict clients:  dict name:data
    :param list products: list with product ids
    :param bool force: forces reinstall if the installed version is the current one
    :param bool now: start installation immediately if possible, otherwise at shutdown
    """

    logger.debug("ui.install: installing products '%s' on clients '%s'" % (products, clients))
    o.install(clients, products, force, now)


def uninstall(o, clients, products, now=False):
    """Remove products from clients.

    :param Opus o: Opus instance
    :param dict clients: dict name:data
    :param list products: list with products ids
    :param bool now: start immediately if possible, otherwise at shutdown
    """

    logger.debug("ui.uninstall: removing products '%s' from clients '%s'" % (products, clients))
    o.uninstall(clients, products, now)


def update(o, clients, products, now=False):
    """Update products on clients.

    Checks all installed products on a client for possible updates and initiates
    an update if one is available.  If the products list is empty, then all
    updatable products will be updated.

    :param Opus o: - Opus instance
    :param dict clients: dict with name:data
    :param bool now: start immediately if possible, otherwise at shutdown
    """

    logger.debug("ui.update: updating products '%s' on clients '%s'" % (products, list(clients)))
    o.update(clients, products, now)


def ondemand(o, clients):
    """Trigger an on-demande event on clients.

    :param Opus o: Opus instance
    :param list clients: list with client ids
     """

    logger.debug("ui.ondemand: Executing ondemand event on '%s'" % clients)
    o.ondemand(clients)


def shutdown(o, clients):
    """Shutdown clients.

    :param Opus o: Opus instance
    :param list clients: list with client ids
    """

    logger.debug("ui.shutdown: shutting down '%s'" % clients)
    o.shutdown(clients)


def reboot(o, clients):
    """Reboot clients.

    :param Opus o: Opus instance
    :param list clients: list with client ids
    """

    logger.debug("ui.reboot: rebooting '%s'" % clients)
    o.reboot(clients)


def list_(clients):
    """Print list with client infos.

    :param dict clients: dictionary with name:values dict
    """

    logger.debug("ui.list_: listing clients")
    if len(clients) == 0:
        click.echo(f"If you are looking for group members, then use 'opus N@groupname members'.")
        return

    for name, json in clients.items():
        ip_address = json["ipAddress"] or "-"
        last_seen = json["lastSeen"] or "-"
        if csv_output:
            click.echo(f"{name};{ip_address};{last_seen}")
        else:
            click.echo(f"{name: <51}  {ip_address: <15} {last_seen}")


def details(o, clients):
    """Print client list with more details (installed products).

    :param: o - Opus instance
    :param: clients - list with client ids
    """

    logger.debug("details for clients")
    list_updatable(o, clients, False)

def os_version(o, clients):
    """Print list with clients and their OS version.

    :param o: Opus instance
    :param clients: list with client ids
    """

    for name, json in clients.items():
        if version := o.os_version(name):
            if csv_output:
                click.secho(f"{name};{version}")
            else:
                click.secho(f"{name: <51}: {version}")
        else:
            click.secho(f"{name: <51}: No software audit.")

def list_updatable(o, clients, brief=True):
    """Print list with clients with possible updates.

    :param o: Opus instance
    :param clients: list with client ids
    :param brief: short version output
    """

    # needed for `details`
    if brief:
        logger.debug("list updatable products")

    products_in_depot = o.products()
    for name, json in clients.items():
        ip_address = json["ipAddress"] or "-"
        last_seen = json["lastSeen"] or "-"
        if csv_output:
            click.echo(f"{name};{ip_address}:{last_seen}")
        else:
            click.echo(f"{name: <50}  {ip_address: <15} {last_seen}")

        logger.debug("json data: %s" % json)
        # get products installed on the client and check for updates
        for product_triple in json["products"]:
            id_ = product_triple[0]
            # ignore possibly dangerous package
            if id_ == 'windomain':
                click.secho(f"  {id_: <30} ignored!", fg="yellow")
                continue
            version = product_triple[1]
            package_version = product_triple[2]

            if new_version := o.updatable(product_triple, products_in_depot):
                if csv_output:
                    click.secho(f";{id_};{version}-{package_version};{new_version}")
                else:
                    click.secho(f"  {id_: <30} {version}-{package_version} -> {new_version}", fg="red")
            elif not brief:
                if csv_output:
                    click.secho(f";{id_};{version}-{package_version};")
                else:
                    click.secho(f"  {id_: <30} {version}-{package_version}")

def list_failed(o, clients):
    """Print list with failed client actions (setup, install, uninstall).

    :param: o - Opus instance
    :param: clients - list with client ids
    """

    logger.debug("listing failed clients")
    failed_clients = o.failedClients(clients)
    for name, products in failed_clients.items():
        if csv_output:
            pass
        else:
            click.echo(f"{name}")

        for product in products:
            if csv_output:
                click.echo(f"{name};{product}")
            else:
                click.echo(f"  {product}")


def list_products(o):
    """List available products (sorted).

    :param: o - Opus instance
    """

    logger.debug("listing products")
    products = o.products()
    for id_, product in sorted(products.items()):
        version = product["productVersion"]
        package_version = product["packageVersion"]
        if csv_output:
            click.echo(f"{id_};{version}-{package_version}")
        else:
            click.echo(f"{id_: <30} {version}-{package_version}")


def list_product_groups(o):
    """List known product groups.

    :param: o - Opus instance
    """

    logger.debug("listing product groups")
    for group in o.productGroups():
        click.echo(f"{group}")


def list_product_group_members(o, group_id):
    """List product group members.

    :param: o - Opus instance
    :param: group_id - name of a group
    """

    logger.debug("listing product group members")
    group_id = "".join(group_id)
    if len(group_id):
        products = o.products()
        for p in o.productGroupMembers(group_id):
            id_ = products[p]["id"]
            version = products[p]["productVersion"]
            package_version = products[p]["packageVersion"]
            click.echo(f"{id_: <30} {version}-{package_version}")
    else:
        click.echo("no group id given")


def list_groups(o, group_id):
    """List client sub groups.

    :param: o - Opus instance
    :param: group_id - name of a group
    """
    logger.debug("listing client groups")

    group_id = "".join(group_id)
    for k, v in o.clientGroups(group_id).items():
        click.echo(f"{k}, {v}")


def list_group_members(o, group_id, clients):
    """List members of a group.

    :param: o - Opus instance
    :param: group_id - name of a group
    :param: clients - clients
    """

    logger.debug("ui:listing client group members for '%s'" % group_id)
    if len(group_id):
        group_id = "".join(group_id) # why?
        # check if group exist in all groups list
        if group_id in o.clientGroups("") or group_id=="":
            # get all clients for a group
            clients = o.clientGroupMembers(group_id)
            for c in clients:
                click.echo("%s" % c)
    else:
        click.echo("no group id given")


def add_client_to_group(o, clients, group_id):
    """Add client(s) to a group.

    :param: o - Opus instance
    :param: clients - dictionary with name:values dict
    :param: group_id - name of a group
    """

    logger.debug("ui.add_client_to_group: adding clients %s to group %s" % (list(clients), group_id))
    if len(group_id)>0 and len(clients)>0:
        group_id = "".join(group_id)
        # check if group exist in the all groups list
        if group_id in o.clientGroups(""):
            o.addClientsToGroup(list(clients), group_id)
    else:
        click.echo("no group id given or empty client list")


def list_unpackaged(o, clients):
    """List unpackaged software.

    List all installed software that is *not* managed via opsi.

    :param: o - Opus instance
    :param: clients - dict name:values
    """

    logger.debug("listing unpackaged software")
    for client, json in clients.items():
        click.echo(client)
        unpackaged_sw = o.unpackaged(client, json)
        for s in unpackaged_sw:
            click.echo(f"{s}")


def list_migratable(o, clients):
    """List installed software available as opsi packages for every given client.

    :param: o - Opus instance
    :param: clients - dict client:data
    """

    logger.debug("ui.listing_migratable")
    for client, json in clients.items():
        available_packages, missing_packages= o.list_migratable(client, json)

        installed_packages = []
        migratable_packages = []

        click.echo(client)
        if len(available_packages) > 0:
            click.echo("Found software that exists as opsi packages:")
            for p in sorted(available_packages):
                    click.echo(f"  {p}")

        if len(missing_packages) > 1:
            click.echo("Found software without an opsi package:")
            for p in sorted(missing_packages):
                click.echo(f"  {p}")


def migrate(o, clients, masked_pkgs):
    """Migrate clients to opsi.

    :param: o - Opus instance
    :param: clients - list with client ids
    """

    logger.debug("migrating clients")
    o.migrate(clients, masked_pkgs)


def list_fresh_clients(o, clients):
    """List 'fresh' clients.

    Print a list with new installed clients (only opsi-script and the client
    agent are installed).

    :param: o - Opus instance
    :param: clients - list with client ids
    """

    logger.debug("ui.list_fresh_clients")
    for name, json in clients.items():
        if o.is_fresh(json):
            click.echo(f"{name}")


def set_installed(o, clients, products, installed=True):
    """Set products to 'installed'.

    :param: o -  Opus instance
    :param: clients - dict name:data
    :param: products - list with product ids
    """
    logger.debug("ui.set_installed (installed=%s)" % installed)

    o.set_installed(list(clients), products, installed)


@click.command()
@click_logging.simple_verbosity_option(logger)
@click.version_option(version=__version__, prog_name=__prog_name__)
@click.option("-c", "--conf", default="", help="path to config file")
@click.option("-i", "--info", is_flag=True, help="print information about server connection and exit")
@click.option("--now", is_flag=True, default=False, help="start action immediately")
@click.option("-f", "--force", is_flag=True, help="force installation")
@click.option("-d", "--depot", default="", help="depot name; use -i for a list of configured depots")
@click.option("-s", "--csv", is_flag=True, help="print output as as CSV (separator is ;)")
@click.argument("arguments", nargs=-1)
def main(conf, info, arguments, now, force, depot, csv):
    """Run ACTION on CLIENTS (with PRODUCTS)

    CLIENTS: specify clients to act on.  Default mode is globbing, e.g. '*'.

      E@uwiw-.* : regular expression

      N@uwiw-group: named group of clients

      L@a,b: list of clients

    ACTION:

        - install/uninstall/update:  install, uninstall or update
          product(s). Use `--now` to start the action immediately.  Use
          `--force` to force an installation even if the version of the product
          is already installed.\n
        - failed: list clients with failed actions (setup, update, uninstall)
        - depotUpdates: list available updates on depot (e.g. from o4i repos)\n
        - shutdown/reboot: shutdown or reboot client(s) immediately\n
        - list/details: show client(s) overview or detailed information\n
        - updatable: show client(s) with possible updates\n
        - groups: list known groups\n
        - members: list clients of a given group\n
        - addtogroup: add clients to a group\n
        - ondemand: trigger a ondemand event, e.g. to start a previously set
          install, uninstall or update action\n
        - productgroups: list known product groups\n
        - productgroupmembers <group>: list members of a product group\n
        - products|packages: list available products\n
        - unpackaged: list installed, but unpackaged software\n
        - fresh: list fresh (=without additional products) installed clients\n
        - set_installed: set state of product(s) to 'installed'\n
        - set_notinstalled: set state of product(s) to 'not_installed'\n
        - migratable: list installed software with available opsi products\n
        - migrate: set products for installed software to 'installed'
        - os_version: print client and its os version

    PRODUCTS: list of products to work with

    Example: Install the product `firefox` and `thunderbird` on all known clients:

        opus --now '*' install firefox thunderbird

    """

    # define output format for printing actions
    global csv_output
    csv_output = csv

    logger.debug("starting opus")

    # TODO: multiplex depots!
    url, username, password, masked_pkgs = get_credentials(conf)

    # print info and exit
    if info:
        print_information(url, username)
        sys.exit()

    # check for allowed actions
    commands = [
            "install",                  # install package
            "uninstall",                # uninstall package
            "update",                   # install if update is possible
            "failed",                   # list failed clients
            "shutdown",                 # shutdown client(s)
            "reboot",                   # reboot client(s)
            "list",                     # simple client list
            "details",                  # client list with package details
            "ondemand",                 # start all previously set actions
            "packages",                 # list available packages
            "products",                 # list available packages
            "groups",
            "productgroups",
            "productgroupmembers",
            "members",
            "addtogroup",
            "depotUpdates",	            # available updates for depot
            #"installDepotUpdates",      # install available updates for depot
            "updatable",                # list clients with possible updates
            "unpackaged",               # list unpackaged software
            "fresh",                    # list fresh installed clients
            "set_installed",
            "set_notinstalled",
            "migratable",
            "migrate",
            "os_version",
            ]

    if len(arguments) < 2:
        click.echo("Missing client and/or action argument.")
        click.echo("Known actions are: %s" % " ".join(list(commands)))
        sys.exit(5)
    clients = arguments[0]
    action = arguments[1]
    if len(arguments) > 2:
        packages = arguments[2:]
    else:
        packages = []

    # pre-check available commands
    if not action in commands:
        click.echo("Unknown action: %s." % action)
        click.echo("Known actions are: %s" % " ".join(list(commands)))
        sys.exit(3)

    # prepare main instance
    try:
        logger.debug("autenticating...")
        o = app.Opus(url, username, password)
        logger.debug("autenticated")
    except AuthenticationError as err:
        click.echo("Authentication failure.", err=True)
        sys.exit(1)
    except ConnectionError as err:
        click.echo("Cannot open a connection to URL '%s'." % url, err=True)
        sys.exit(2)

    # what do we have? clients/groups or depot?
    # create clients list
    if clients:
        clients_ = o.evaluate_targets(clients)
        if len(clients_)==0:
            click.echo("No clients or groups found with '%s'.  Exiting." % clients, err=True)
            sys.exit()
        else:
            clients = clients_
    logger.debug("clients: %s" % clients)

    # make list from packages parameter string
    if packages:
        packages = list(packages)

    if action=="install":
        install(o, clients, packages, force, now)
    elif action=="uninstall":
        uninstall(o, clients, packages, now)
    elif action=="update":
        update(o, clients, packages, now)
    elif action=="failed":
        list_failed(o, clients)
    elif action=="shutdown":
        shutdown(o, clients)
    elif action=="reboot":
        reboot(o, clients)
    elif action=="list" or action=="members":
        # the members part is handled by the targeting string
        list_(clients)
    elif action=="details":
        details(o, clients)
    elif action=="updatable":
        list_updatable(o, clients)
    elif action=="ondemand":
        ondemand(o, clients)
    elif action=="packages" or action=="products":
        list_products(o)
    elif action=="groups":
        list_groups(o, packages)
    elif action=="addtogroup":
        add_client_to_group(o, clients, group_id=packages)
    elif action=="productgroups":
        list_product_groups(o)
    elif action=="productgroupmembers":
        list_product_group_members(o, packages)
    #elif action=="depotUpdates":
    #    list_available_updates(o, ssh_cmd)
    #elif action=="installDepotUpdates":
    #    install_updates_on_depot(o, ssh_cmd)
    elif action=="unpackaged":
        list_unpackaged(o, clients)
    elif action=="fresh":
        list_fresh_clients(o, clients)
    elif action=="set_installed":
        set_installed(o,clients, packages, True)
    elif action=="set_notinstalled":
        set_installed(o, clients, packages, False)
    elif action=="migratable":
        list_migratable(o, clients)
    elif action=="migrate":
        migrate(o, clients, masked_pkgs)
    elif action=="os_version":
        os_version(o, clients)
    else:
        click.echo("unknown actions; this should not happen!", err=True)
        sys.exit(3)


if __name__ == "__main__":
    main()
